#!/bin/bash

openssl genrsa -aes-256-cbc -out test_private.pem 2048
openssl rsa -in test_private.pem -pubout > test_public.pem
touch test_rsa_native_results.csv

for INDEX in {1..1000}
do
    echo Iteration $INDEX;
    cat /dev/urandom | tr -dc '[:alpha:]' | fold -w ${1:-16000000} | head -n 1 > random.txt;
    echo -n $INDEX >> test_rsa_native_results.csv;
    /usr/bin/time -a -o test_rsa_native_results.csv -f ',ENC,%E,%U,%S,%M,%K,%D,%X' openssl enc -aes-256-cbc -in random.txt -out top_secret.enc -pbkdf2 -pass file:test_private.pem -kfile mypass.txt;
    echo -n $INDEX >> test_rsa_native_results.csv;
    /usr/bin/time -a -o test_rsa_native_results.csv -f ',DEC,%E,%U,%S,%M,%K,%D,%X' openssl enc -d -aes-256-cbc -in top_secret.enc -out top_secret.out -pbkdf2 -pass file:test_private.pem -kfile mypass.txt;
    # Verify Decryption is successful
    VARA=$(md5sum random.txt | cut -d ' ' -f 1);
    VARB=$(md5sum top_secret.out | cut -d ' ' -f 1);
    if [ $VARA != $VARB ]; then
        echo 'Failed to decrypt file';
        echo $VARA;
        echo $VARB;
        exit;
    fi
done

rm -f random.txt top_secret.enc top_secret.out test_private.pem test_public.pem
