#!/home/vagrant/janus/venv/bin/python3
"""Webserver

This module is ment to connect client requests and will
manage the CURD operations to the private and csp backend

Example:
    $ python webserver.py --host 0.0.0.0 --port 80

Attributes:
    host (str):
    port (int):

TODO:
    US.TIK.1-3
    US.CIK.1-4
    US.Create.1-9
    US.Update.1-6
    US.Read.1-4
    US.Delete.1-3
    US.Search.1-14

Sources:
https://flask.palletsprojects.com/en/2.0.x/patterns/fileuploads/
https://boto3.amazonaws.com/v1/documentation/api/latest/guide/quickstart.html
https://subsystic.com/documentation/id-secret-access-key-amazon-s3/
"""
import os
import argparse
import logging
import subprocess
import threading
import shutil
import json
import configparser
import boto3
import random
import string
import time
import requests
import pymysql
from pymysql.constants import CLIENT
import common
from flask import Flask, request, jsonify, make_response, flash, url_for, send_from_directory
from werkzeug.utils import secure_filename
from functools import wraps
from secretpy import ColumnarTransposition, alphabets as al

HOST_IP = ''
PORT = ''
UPLOAD_FOLDER = common.UPLOAD_FOLDER
ALLOWED_EXTENSIONS = common.ALLOWED_EXTENSIONS
CONFIG_PARSER = configparser.ConfigParser()
conn = None
s3_private_client = None
s3_csp_client = None
OPENABE_HOSTNAME = ''
OPENABE_PORT = ''

logging.basicConfig(format='%(asctime)s %(message)s')
logger = logging.getLogger('webserver')

def check_auth(usr, password):
    # TODO: auth user
    return usr == 'test' and password == 'password'

def login_required(f):
    @wraps(f)
    def wrapped_view(**kwargs):
        auth = request.authorization
        if not (auth and check_auth(auth.username, auth.password)):
            return ('Unauthorized', 401, {
                'WWW-Authenticate': 'Basic realm="Login Required"'
            })
        return f(**kwargs)
    return wrapped_view

app = Flask(__name__)
app.config['UPLOAD_FOLDER'] = UPLOAD_FOLDER
app.config['MAX_CONTENT_LENGTH'] = 16 * 1000 * 1000 #16 MB

@app.route("/")
def hello_world():
    return "<p>Hello World<p>"

# Public routes
@app.route("/create", methods=['POST'])
def create():
    """Create new resource

    Args:
        username: (request.authorization.username)
        password: (request.authorization.password)
        file (file): File content
        org (str): Org MSK used or to use
        key (str): Key encypted with or to use
        kpabe_encrypted (bool): weather the files is already encypted
        word_set_extraction (:obj:`list` of :obj: `str`, optional)

    Returns:
        :obj:`request.response` of common.message \
            and :obj:`str` of HTTP /response/<siid>

    Codes:
        200 Success
        400 Bad_Request
        413 Payload_Too_Long
        500 Internal_Server_Error

    Example:
        curl -s -X POST -F "file=@hello_world.txt" -F "json=@data.json" http://192.168.1.4:8080/create | python3 -m json.tool
    """
    m_start = time.time()
    dcontent = ''
    fcontent = ''
    if 'json' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `json`')
    dfile = request.files['json']
    if dfile and common.allowed_file(dfile.filename):
        dfilename = secure_filename(dfile.filename)
        dfile.save(os.path.join(app.config['UPLOAD_FOLDER'], dfilename))
    with open(os.path.join(app.config['UPLOAD_FOLDER'], dfile.filename)) as f:
        dcontent = f.readlines()
    data = json.loads(''.join(dcontent))
    verify_key_data(data)
    # check if the post request has the file part
    file = verify_secure_file('file', request.files)
    kpabe_encrypted = data.get('kpabe_encrypted', None)
    if kpabe_encrypted is None:
        return common.default_error_code(400, 'kpabe_encrypted required', 'Client JSON Error')
    word_set_extraction = data.get('word_set_extraction', None)
    if word_set_extraction is None:
        cipher = ColumnarTransposition()
        alphabet = al.ENGLISH
        plaintext = u"{0}".format('')
        key = u"{0}".format('key')
        #print(plaintext)
        enc = cipher.encrypt(plaintext, key, alphabet)
        #print(enc)
    if kpabe_encrypted:
        if word_set_extraction == false:
            return common.default_error_code(400, 'word_set_extraction required', 'when kpabe_encrypted is true word_set_extraction is required')

    with open(os.path.join(app.config['UPLOAD_FOLDER'], file.filename)) as f:
        fcontent = f.readlines()

    #Generate IDs For Cloud Storage
    random_private_filename = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '.txt'
    random_csp_filename = ''.join(random.choice(string.ascii_lowercase) for i in range(10)) + str(int(time.time()*1000.0)) + '.txt'
    #tmp_input_file = os.path.join(app.config["UPLOAD_FOLDER"], random_input)

    # Get SIID from data owner SECURE_INDEX_ID
    m_query = "INSERT INTO siid (siid) VALUES (NULL);"
    m_cursor = conn.cursor()
    m_cursor.execute(m_query)
    m_cursor.execute("SELECT MAX(siid) FROM siid;")
    results = m_cursor.fetchall()
    siid = results[0][0]
    conn.commit()

    # Set prototype_search_ingest
    #m_query = "INSERT INTO prototype_search (siid, extracted_text) VALUES ({0}, '{1}');".format(siid, ''.join(fcontent))
    #m_cursor = conn.cursor()
    #m_cursor.execute(m_query)
    #conn.commit()

    # Set plain_text_keys
    m_query = "INSERT INTO plain_text_keys (siid, s3_plain_text_key, s3_cypher_text_key, msk, mpk) VALUES ({0}, \'{1}\', \'{2}\', \'{3}.msk\', \'{3}.mpk\');".format(siid, random_private_filename, random_csp_filename, data['org'])
    m_cursor = conn.cursor()
    m_cursor.execute(m_query)
    conn.commit()

    # Set secure_encrypted_index
    m_query = "INSERT INTO secure_encrypted_index (siid, s3_cypher_text_key) VALUES ({0}, \'{1}\');".format(siid, random_csp_filename)
    m_cursor = conn.cursor()
    m_cursor.execute(m_query)
    conn.commit()

    # Save S3 Private
    m_start_s3_private_client = time.time()
    response = s3_private_client.put_object(Body=''.join(fcontent), Key=random_private_filename, Bucket='private')
    try:
        if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
            return common.default_error_code(500, 's3_private.response', response)
    except Exception as e:
        return common.default_error_code(500, 's3_private.exception', e)
    m_end_s3_private_client = time.time()
    # Create Event Return Code

    # Document Encryption if not encrypted
    # https://docs.python-requests.org/en/latest/user/advanced/#post-multiple-multipart-encoded-files
    m_start_kpabe_encrypted = time.time()
    if not kpabe_encrypted:
        #print('kpabe_encrypted:')
        #print(type(request.files['file']))
        #print(request.files['file'])
        #print(fcontent)
        multiple_files = {
            'json': open(os.path.join(app.config['UPLOAD_FOLDER'], dfilename), 'rb'),
            'file': open(os.path.join(app.config['UPLOAD_FOLDER'], file.filename), 'rb')
        }
        #print('request')
        r = requests.get('http://{0}:{1}/encrypt'.format(OPENABE_HOSTNAME, OPENABE_PORT), auth=('test', 'password'), files=multiple_files)
        if r.status_code != 200:
            return common.default_error_code(r.status_code, '{0}'.format(r.content), r.reason)
    m_end_kpabe_encrypted = time.time()

    # Upload to S3 csp public
    m_start_s3_csp_client = time.time()
    response = s3_csp_client.put_object(Body=r.content, Key=random_csp_filename, Bucket='public')
    try:
        if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
            return common.default_error_code(500, 's3_csp.response', response)
    except Exception as e:
        return common.default_error_code(500, 's3_csp.exception', e)
    if os.path.exists(os.path.join(app.config['UPLOAD_FOLDER'], dfile.filename)):
        os.remove(os.path.join(app.config['UPLOAD_FOLDER'], dfile.filename))
    if os.path.exists(os.path.join(app.config['UPLOAD_FOLDER'], file.filename)):
        os.remove(os.path.join(app.config['UPLOAD_FOLDER'], file.filename))
    m_end_s3_csp_client = time.time()
    with open('test_webserver_create.csv', 'a') as f:
        f.write('{0},{1},{2},{3}\n'.format(m_end_s3_csp_client-m_start,
                                         m_end_kpabe_encrypted-m_start_kpabe_encrypted,
                                         m_end_s3_private_client-m_start_s3_private_client,
                                         m_end_s3_csp_client-m_start_s3_csp_client));
    # Update Event Return Code
    return common.success_code_with_extra_data({'response_resource':
                                                '{0}://{1}:{2}/{3}/{4}'.format('http', HOST_IP, PORT, 'responses', siid)})

@app.route("/update/<siid>")
def update(siid):
    m_query = "DELETE FROM secure_encrypted_index WHERE siid = %s;"
    m_cursor = conn.cursor()
    m_cursor.execute(m_query, (siid))
    conn.commit()
    return "TODO: update"

@app.route("/search", methods=['POST'])
def search():
    print('test search')
    if 'json' not in request.files:
        return common.default_error_code(400, 'form-data', 'no file content in `json`')
    data = json.loads(request.files['json'].read())
    if data is None:
        return common.default_error_code(400, 'form-data', 'json content empty')
    s3_files = data.get('s3_files', False)
    print(s3_files)
    print(type(s3_files))
    list_words = data.get('list_words', None)
    if list_words is None:
        return common.default_error_code(400, 'json list of words', 'words needed for search')
    # Get ResponsesID
    m_query = "INSERT INTO responses (resp_id, message, status_code, stdout, stderr) VALUES (NULL, '{0}', '{1}', '', '');".format('STARTED', 200)
    m_cursor = conn.cursor()
    m_cursor.execute(m_query)
    m_cursor.execute("SELECT MAX(resp_id) FROM responses;")
    results = m_cursor.fetchall()
    responseID = results[0][0]
    conn.commit()

    thread = threading.Thread(target=run_search, args=(responseID, s3_files, ['hello']))
    thread.start()

    return common.success_code_with_extra_data({'response_resource':
                                                '{0}://{1}:{2}/{3}/{4}'.format('http', HOST_IP, PORT, 'responses', responseID)})

@app.route("/read/<siid>", methods=['GET', 'POST'])
def read(siid):
    """Read existing resource

    Args:
        username: (request.authorization.username)
        password: (request.authorization.password)
        siid (str): SECURE_INDEX_ID

    Returns:
        :obj:`request.response` file

    Codes:
        200 Success
        400 Bad_Request
        413 Payload_Too_Long
        500 Internal_Server_Error

    Example:
        curl -s -X POST -F "file=@hello_world.txt" -F "json=@data.json" http://192.168.1.4:8080/create | python3 -m json.tool
    """
    file = None
    mpk = None
    data = None

    if request.method == 'POST':
        file = verify_secure_file('file', request.files)

    m_query = "SELECT s3_cypher_text_key FROM secure_encrypted_index WHERE siid = %s;"
    m_cursor = conn.cursor()
    m_cursor.execute(m_query, (siid))
    results = m_cursor.fetchall()
    f_name = results[0][0]

    response = s3_csp_client.get_object(Key=f_name, Bucket='public')
    try:
        if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
            return common.default_error_code(500, 's3_private.response', response)
    except Exception as e:
        return common.default_error_code(500, 's3_private.exception', e)

    content = response.get('Body').read()

    if request.method == 'POST':
        # COPY from create
        if 'json' not in request.files:
            return common.default_error_code(400, 'form-data', 'no file content in `json`')
        dfile = request.files['json']
        if dfile and common.allowed_file(dfile.filename):
            dfilename = secure_filename(dfile.filename)
            dfile.save(os.path.join(app.config['UPLOAD_FOLDER'], dfilename))
        with open(os.path.join(app.config['UPLOAD_FOLDER'], dfile.filename)) as f:
            dcontent = f.readlines()
        data = json.loads(''.join(dcontent))
        verify_key_data(data)

        print('TODO: encrypted document')
        print(type(request.files['file']))
        print(request.files['file'])
        multiple_files = {
            'json': open(os.path.join(app.config['UPLOAD_FOLDER'], dfile.filename), 'rb'),
            'file': open(os.path.join(app.config['UPLOAD_FOLDER'], file.filename), 'rb')
        }
        r = requests.get('http://{0}:{1}/decrypt'.format(OPENABE_HOSTNAME, OPENABE_PORT), auth=('test', 'password'), files=multiple_files)
        if r.status_code != 200:
            return common.default_error_code(r.status_code, '{0}'.format(r.content), r.reason)
        return r.content
    else:
        return content

@app.route("/delete/<siid>")
def delete(siid):
    """Responses to requests

    Args:
        username: (request.authorization.username)
        password: (request.authorization.password)
        siid (str): SECURE_INDEX_ID

    Returns:
        request.response: 200 Success,

    Codes:
        200
        400
        401
        403
        404
        500
    """
    m_query = "DELETE FROM secure_encrypted_index WHERE siid = %s;"
    m_cursor = conn.cursor()
    m_cursor.execute(m_query, (siid))
    conn.commit()
    return common.default_success_code()

# Private Routes
@app.route("/responses/<resp_id>")
def responses(resp_id):
    """Responses to requests

    Args:
        username: (request.authorization.username)
        password: (request.authorization.password)
        siid (str): SECURE_INDEX_ID

    Returns:
        request.response: 200 Success with path to response,

    Codes:
        200
        400
        401
        403
        404
        500
    """
    rtn = {}
    m_cursor = conn.cursor()
    m_query = "SELECT message, status_code, stdout, stderr FROM responses WHERE resp_id = {0};".format(resp_id)
    print(m_query)
    m_cursor.execute(m_query)
    res = [x for x in m_cursor.fetchall()]
    print(res)
    try:
        rtn['response_status'] = res[0][0]
        rtn['status_code'] = res[0][1]
        rtn['stdout'] = res[0][2]
        rtn['stderr'] = res[0][3]
    except:
        rtn['response_status'] = 'Not Found'
        rtn['status_code'] = 404
        rtn['stdout'] = ''
        rtn['stderr'] = ''

    m_query = "SELECT response_url FROM response_urls WHERE resp_id = {0};".format(resp_id)
    print(m_query)
    m_cursor.execute(m_query)
    response_urls = [x[0] for x in m_cursor.fetchall()]
    send_to_aws = False
    if len(response_urls) > 0:
        if not response_urls[0].isdigit():
            send_to_aws = True
    if send_to_aws:
        rtn['response_resources'] = ['{0}/{1}/{2}'.format(CONFIG_PARSER.get("s3_csp_config", "endpoint_url"), 'public', url) for url in response_urls]
    else:
        rtn['response_resources'] = ['{0}://{1}:{2}/{3}/{4}'.format('http', HOST_IP, PORT, 'read', url) for url in response_urls]

    m_cursor.close()
    return common.success_code_with_extra_data(rtn)

def run_search(resp_id, s3_files, list_words):
    # Set prototype_search_ingest
    reponse_code = 200
    stdout = ''
    stderr = ''
    m_cursor = conn.cursor()
    try:
        m_query = "SELECT siid FROM prototype_search WHERE {0};".format(
            ' OR '.join(["(CONVERT(extracted_text USING utf8) LIKE \'%{0}%\')".format(x) for x in list_words]))
        print(m_query)
        m_cursor.execute(m_query)
        siids = [x[0] for x in m_cursor.fetchall()]
        if s3_files:
            m_query = "SELECT s3_cypher_text_key FROM secure_encrypted_index WHERE {0};".format(
                ' OR '.join(["siid = {0}".format(x) for x in siids]))
            print(m_query)
            m_cursor.execute(m_query)
            cipher_keys = [x[0] for x in m_cursor.fetchall()]
            m_query = "INSERT INTO response_urls (resp_id, response_url) VALUES {0};".format(', '.join(["({0}, \'{1}\')".format(resp_id, x) for x in cipher_keys]))
        else:
            m_query = "INSERT INTO response_urls (resp_id, response_url) VALUES {0};".format(', '.join(["({0}, {1})".format(resp_id, x) for x in siids]))
        print(m_query)
        m_cursor.execute(m_query)
        conn.commit()
    except Exception as e:
        reponse_code = 500
        print(e)
    # Update Response Code
    attr = []
    attr.append("message = '{0}'".format('COMPLETE'))
    attr.append("status_code = '{0}'".format(reponse_code))
    if stdout != '':
        attr.append("stdout = '{0}'".format(stdout))
    if stderr != '':
        attr.append("stderr = '{0}'".format(stderr))
    m_query = "UPDATE responses SET {0} WHERE resp_id = {1};".format(', '.join(attr), resp_id)
    m_cursor.execute(m_query)
    conn.commit()

def verify_key_data(data):
    if data is None:
        return common.default_error_code(400, 'form-data', 'json content empty')
    org = data.get('org', None)
    if org is None:
        return common.default_error_code(400, 'org required', 'Client JSON Error')
    key = data.get('key', None)
    if key is None:
        return common.default_error_code(400, 'key required', 'Client JSON Error')
    policy = data.get('policy', None)
    if policy is None:
        return default_error_code(400, 'policy="policy_required', 'Client JSON Error')

def verify_secure_file(name, files_list):
    if name not in files_list:
        return common.default_error_code(400, 'form-data', 'no file content in `{0}`'.format(name))
    file = files_list[name]
    # If the user does not select a file, the browser submits an
    # empty file without a filename.
    if file.filename == '':
        return common.default_error_code(400, 'form-data', 'file content empty in `{0}`'.format(name))
    if file and common.allowed_file(file.filename):
        filename = secure_filename(file.filename)
        file.save(os.path.join(app.config['UPLOAD_FOLDER'], filename))
    return file

def remove_tmp_files(list_tmp_file):
    for f in list_tmp_file:
        remove_tmp_file(f)

def remove_tmp_file(tmp_file):
    tmp_file_path = os.path.join(app.config["UPLOAD_FOLDER"], tmp_file)
    if os.path.exists(tmp_file_path):
        os.remove(tmp_file_path)

def main():
    global conn, HOST_IP, PORT, s3_private_client, s3_csp_client, OPENABE_HOSTNAME, OPENABE_PORT
    parser = argparse.ArgumentParser(description='Webserver to hadle client requests')
    parser = common.common_parser(parser)
    args = parser.parse_args()
    HOST_IP = args.host_ip
    PORT = args.port
    CONFIG_PARSER.read(args.config)

    sql_hostname = CONFIG_PARSER.get("mysql_config", "hostname")
    sql_port = CONFIG_PARSER.get("mysql_config", "port")
    sql_username = CONFIG_PARSER.get("mysql_config", "username")
    sql_dbname = CONFIG_PARSER.get("mysql_config", "database")
    sql_pass = CONFIG_PARSER.get("mysql_config", "password")

    conn = pymysql.connect(host=sql_hostname,
        user=sql_username,
        password=sql_pass,
        db=sql_dbname,
        port=int(sql_port),
        client_flag=CLIENT.MULTI_STATEMENTS)

    s3_private_key_id = CONFIG_PARSER.get("s3_private_config", "aws_access_key_id")
    s3_private_access_key = CONFIG_PARSER.get("s3_private_config", "aws_secret_access_key")
    s3_private_endpoint_url = CONFIG_PARSER.get("s3_private_config", "endpoint_url")

    s3_private_client = boto3.client(
        's3',
        aws_access_key_id=s3_private_key_id,
        aws_secret_access_key=s3_private_access_key,
        endpoint_url=s3_private_endpoint_url
    )

    s3_csp_key_id = CONFIG_PARSER.get("s3_csp_config", "aws_access_key_id")
    s3_csp_access_key = CONFIG_PARSER.get("s3_csp_config", "aws_secret_access_key")
    s3_csp_endpoint_url = CONFIG_PARSER.get("s3_csp_config", "endpoint_url")

    s3_csp_client = boto3.client(
        's3',
        aws_access_key_id=s3_csp_key_id,
        aws_secret_access_key=s3_csp_access_key,
        endpoint_url=s3_csp_endpoint_url
    )

    OPENABE_HOSTNAME = CONFIG_PARSER.get("openabe_wrapper", "hosthame")
    OPENABE_PORT = CONFIG_PARSER.get("openabe_wrapper", "port")

    if conn is None:
        print("Error connecting to mysql database")
    else:
        print("Database connection established")
        print(conn)
    try:
        app.run(host=args.host_ip, port=args.port)
    finally:
        print("Connection Close")
        conn.close()

if __name__ == '__main__':
    main()
